/*
 * Copyright (C) 2012 Red Hat Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *     * Redistributions of source code must retain the above
 *       copyright notice, this list of conditions and the
 *       following disclaimer.
 *     * Redistributions in binary form must reproduce the
 *       above copyright notice, this list of conditions and
 *       the following disclaimer in the documentation and/or
 *       other materials provided with the distribution.
 *     * The names of contributors to this software may not be
 *       used to endorse or promote products derived from this
 *       software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 * Author: Stef Walter <stefw@redhat.com>
 */

#ifndef P11_TOKEN_H_
#define P11_TOKEN_H_

#include "dict.h"
#include "index.h"
#include "parser.h"
#include "pkcs11.h"

enum {
	P11_TOKEN_FLAG_NONE = 0,
	P11_TOKEN_FLAG_WRITE_PROTECTED = 1 << 0,
};

typedef struct _p11_token p11_token;

p11_token *     p11_token_new         (CK_SLOT_ID slot,
                                       const char *path,
                                       const char *label,
                                       int flags);

void            p11_token_free        (p11_token *token);

int             p11_token_load        (p11_token *token);

bool            p11_token_reload      (p11_token *token,
                                       CK_ATTRIBUTE *attrs);

p11_index *     p11_token_index       (p11_token *token);

p11_parser *    p11_token_parser      (p11_token *token);

const char *    p11_token_get_path    (p11_token *token);

const char *    p11_token_get_label   (p11_token *token);

CK_SLOT_ID      p11_token_get_slot    (p11_token *token);

bool            p11_token_is_writable (p11_token *token);

#endif /* P11_TOKEN_H_ */
