P11_VIRTUAL_FIXED_FUNCTIONS(0)
P11_VIRTUAL_FIXED_FUNCTIONS(1)
P11_VIRTUAL_FIXED_FUNCTIONS(2)
P11_VIRTUAL_FIXED_FUNCTIONS(3)
P11_VIRTUAL_FIXED_FUNCTIONS(4)
P11_VIRTUAL_FIXED_FUNCTIONS(5)
P11_VIRTUAL_FIXED_FUNCTIONS(6)
P11_VIRTUAL_FIXED_FUNCTIONS(7)
P11_VIRTUAL_FIXED_FUNCTIONS(8)
P11_VIRTUAL_FIXED_FUNCTIONS(9)
P11_VIRTUAL_FIXED_FUNCTIONS(10)
P11_VIRTUAL_FIXED_FUNCTIONS(11)
P11_VIRTUAL_FIXED_FUNCTIONS(12)
P11_VIRTUAL_FIXED_FUNCTIONS(13)
P11_VIRTUAL_FIXED_FUNCTIONS(14)
P11_VIRTUAL_FIXED_FUNCTIONS(15)
P11_VIRTUAL_FIXED_FUNCTIONS(16)
P11_VIRTUAL_FIXED_FUNCTIONS(17)
P11_VIRTUAL_FIXED_FUNCTIONS(18)
P11_VIRTUAL_FIXED_FUNCTIONS(19)
P11_VIRTUAL_FIXED_FUNCTIONS(20)
P11_VIRTUAL_FIXED_FUNCTIONS(21)
P11_VIRTUAL_FIXED_FUNCTIONS(22)
P11_VIRTUAL_FIXED_FUNCTIONS(23)
P11_VIRTUAL_FIXED_FUNCTIONS(24)
P11_VIRTUAL_FIXED_FUNCTIONS(25)
P11_VIRTUAL_FIXED_FUNCTIONS(26)
P11_VIRTUAL_FIXED_FUNCTIONS(27)
P11_VIRTUAL_FIXED_FUNCTIONS(28)
P11_VIRTUAL_FIXED_FUNCTIONS(29)
P11_VIRTUAL_FIXED_FUNCTIONS(30)
P11_VIRTUAL_FIXED_FUNCTIONS(31)
P11_VIRTUAL_FIXED_FUNCTIONS(32)
P11_VIRTUAL_FIXED_FUNCTIONS(33)
P11_VIRTUAL_FIXED_FUNCTIONS(34)
P11_VIRTUAL_FIXED_FUNCTIONS(35)
P11_VIRTUAL_FIXED_FUNCTIONS(36)
P11_VIRTUAL_FIXED_FUNCTIONS(37)
P11_VIRTUAL_FIXED_FUNCTIONS(38)
P11_VIRTUAL_FIXED_FUNCTIONS(39)
P11_VIRTUAL_FIXED_FUNCTIONS(40)
P11_VIRTUAL_FIXED_FUNCTIONS(41)
P11_VIRTUAL_FIXED_FUNCTIONS(42)
P11_VIRTUAL_FIXED_FUNCTIONS(43)
P11_VIRTUAL_FIXED_FUNCTIONS(44)
P11_VIRTUAL_FIXED_FUNCTIONS(45)
P11_VIRTUAL_FIXED_FUNCTIONS(46)
P11_VIRTUAL_FIXED_FUNCTIONS(47)
P11_VIRTUAL_FIXED_FUNCTIONS(48)
P11_VIRTUAL_FIXED_FUNCTIONS(49)
P11_VIRTUAL_FIXED_FUNCTIONS(50)
P11_VIRTUAL_FIXED_FUNCTIONS(51)
P11_VIRTUAL_FIXED_FUNCTIONS(52)
P11_VIRTUAL_FIXED_FUNCTIONS(53)
P11_VIRTUAL_FIXED_FUNCTIONS(54)
P11_VIRTUAL_FIXED_FUNCTIONS(55)
P11_VIRTUAL_FIXED_FUNCTIONS(56)
P11_VIRTUAL_FIXED_FUNCTIONS(57)
P11_VIRTUAL_FIXED_FUNCTIONS(58)
P11_VIRTUAL_FIXED_FUNCTIONS(59)
P11_VIRTUAL_FIXED_FUNCTIONS(60)
P11_VIRTUAL_FIXED_FUNCTIONS(61)
P11_VIRTUAL_FIXED_FUNCTIONS(62)
P11_VIRTUAL_FIXED_FUNCTIONS(63)

CK_FUNCTION_LIST_3_0 p11_virtual_fixed[P11_VIRTUAL_MAX_FIXED] = {
        P11_VIRTUAL_FIXED_INITIALIZER(0),
        P11_VIRTUAL_FIXED_INITIALIZER(1),
        P11_VIRTUAL_FIXED_INITIALIZER(2),
        P11_VIRTUAL_FIXED_INITIALIZER(3),
        P11_VIRTUAL_FIXED_INITIALIZER(4),
        P11_VIRTUAL_FIXED_INITIALIZER(5),
        P11_VIRTUAL_FIXED_INITIALIZER(6),
        P11_VIRTUAL_FIXED_INITIALIZER(7),
        P11_VIRTUAL_FIXED_INITIALIZER(8),
        P11_VIRTUAL_FIXED_INITIALIZER(9),
        P11_VIRTUAL_FIXED_INITIALIZER(10),
        P11_VIRTUAL_FIXED_INITIALIZER(11),
        P11_VIRTUAL_FIXED_INITIALIZER(12),
        P11_VIRTUAL_FIXED_INITIALIZER(13),
        P11_VIRTUAL_FIXED_INITIALIZER(14),
        P11_VIRTUAL_FIXED_INITIALIZER(15),
        P11_VIRTUAL_FIXED_INITIALIZER(16),
        P11_VIRTUAL_FIXED_INITIALIZER(17),
        P11_VIRTUAL_FIXED_INITIALIZER(18),
        P11_VIRTUAL_FIXED_INITIALIZER(19),
        P11_VIRTUAL_FIXED_INITIALIZER(20),
        P11_VIRTUAL_FIXED_INITIALIZER(21),
        P11_VIRTUAL_FIXED_INITIALIZER(22),
        P11_VIRTUAL_FIXED_INITIALIZER(23),
        P11_VIRTUAL_FIXED_INITIALIZER(24),
        P11_VIRTUAL_FIXED_INITIALIZER(25),
        P11_VIRTUAL_FIXED_INITIALIZER(26),
        P11_VIRTUAL_FIXED_INITIALIZER(27),
        P11_VIRTUAL_FIXED_INITIALIZER(28),
        P11_VIRTUAL_FIXED_INITIALIZER(29),
        P11_VIRTUAL_FIXED_INITIALIZER(30),
        P11_VIRTUAL_FIXED_INITIALIZER(31),
        P11_VIRTUAL_FIXED_INITIALIZER(32),
        P11_VIRTUAL_FIXED_INITIALIZER(33),
        P11_VIRTUAL_FIXED_INITIALIZER(34),
        P11_VIRTUAL_FIXED_INITIALIZER(35),
        P11_VIRTUAL_FIXED_INITIALIZER(36),
        P11_VIRTUAL_FIXED_INITIALIZER(37),
        P11_VIRTUAL_FIXED_INITIALIZER(38),
        P11_VIRTUAL_FIXED_INITIALIZER(39),
        P11_VIRTUAL_FIXED_INITIALIZER(40),
        P11_VIRTUAL_FIXED_INITIALIZER(41),
        P11_VIRTUAL_FIXED_INITIALIZER(42),
        P11_VIRTUAL_FIXED_INITIALIZER(43),
        P11_VIRTUAL_FIXED_INITIALIZER(44),
        P11_VIRTUAL_FIXED_INITIALIZER(45),
        P11_VIRTUAL_FIXED_INITIALIZER(46),
        P11_VIRTUAL_FIXED_INITIALIZER(47),
        P11_VIRTUAL_FIXED_INITIALIZER(48),
        P11_VIRTUAL_FIXED_INITIALIZER(49),
        P11_VIRTUAL_FIXED_INITIALIZER(50),
        P11_VIRTUAL_FIXED_INITIALIZER(51),
        P11_VIRTUAL_FIXED_INITIALIZER(52),
        P11_VIRTUAL_FIXED_INITIALIZER(53),
        P11_VIRTUAL_FIXED_INITIALIZER(54),
        P11_VIRTUAL_FIXED_INITIALIZER(55),
        P11_VIRTUAL_FIXED_INITIALIZER(56),
        P11_VIRTUAL_FIXED_INITIALIZER(57),
        P11_VIRTUAL_FIXED_INITIALIZER(58),
        P11_VIRTUAL_FIXED_INITIALIZER(59),
        P11_VIRTUAL_FIXED_INITIALIZER(60),
        P11_VIRTUAL_FIXED_INITIALIZER(61),
        P11_VIRTUAL_FIXED_INITIALIZER(62),
        P11_VIRTUAL_FIXED_INITIALIZER(63)
};

P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(0)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(0)
P11_VIRTUAL_FIXED_GET_INTERFACE(0)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(1)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(1)
P11_VIRTUAL_FIXED_GET_INTERFACE(1)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(2)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(2)
P11_VIRTUAL_FIXED_GET_INTERFACE(2)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(3)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(3)
P11_VIRTUAL_FIXED_GET_INTERFACE(3)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(4)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(4)
P11_VIRTUAL_FIXED_GET_INTERFACE(4)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(5)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(5)
P11_VIRTUAL_FIXED_GET_INTERFACE(5)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(6)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(6)
P11_VIRTUAL_FIXED_GET_INTERFACE(6)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(7)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(7)
P11_VIRTUAL_FIXED_GET_INTERFACE(7)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(8)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(8)
P11_VIRTUAL_FIXED_GET_INTERFACE(8)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(9)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(9)
P11_VIRTUAL_FIXED_GET_INTERFACE(9)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(10)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(10)
P11_VIRTUAL_FIXED_GET_INTERFACE(10)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(11)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(11)
P11_VIRTUAL_FIXED_GET_INTERFACE(11)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(12)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(12)
P11_VIRTUAL_FIXED_GET_INTERFACE(12)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(13)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(13)
P11_VIRTUAL_FIXED_GET_INTERFACE(13)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(14)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(14)
P11_VIRTUAL_FIXED_GET_INTERFACE(14)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(15)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(15)
P11_VIRTUAL_FIXED_GET_INTERFACE(15)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(16)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(16)
P11_VIRTUAL_FIXED_GET_INTERFACE(16)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(17)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(17)
P11_VIRTUAL_FIXED_GET_INTERFACE(17)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(18)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(18)
P11_VIRTUAL_FIXED_GET_INTERFACE(18)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(19)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(19)
P11_VIRTUAL_FIXED_GET_INTERFACE(19)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(20)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(20)
P11_VIRTUAL_FIXED_GET_INTERFACE(20)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(21)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(21)
P11_VIRTUAL_FIXED_GET_INTERFACE(21)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(22)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(22)
P11_VIRTUAL_FIXED_GET_INTERFACE(22)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(23)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(23)
P11_VIRTUAL_FIXED_GET_INTERFACE(23)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(24)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(24)
P11_VIRTUAL_FIXED_GET_INTERFACE(24)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(25)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(25)
P11_VIRTUAL_FIXED_GET_INTERFACE(25)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(26)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(26)
P11_VIRTUAL_FIXED_GET_INTERFACE(26)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(27)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(27)
P11_VIRTUAL_FIXED_GET_INTERFACE(27)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(28)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(28)
P11_VIRTUAL_FIXED_GET_INTERFACE(28)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(29)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(29)
P11_VIRTUAL_FIXED_GET_INTERFACE(29)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(30)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(30)
P11_VIRTUAL_FIXED_GET_INTERFACE(30)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(31)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(31)
P11_VIRTUAL_FIXED_GET_INTERFACE(31)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(32)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(32)
P11_VIRTUAL_FIXED_GET_INTERFACE(32)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(33)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(33)
P11_VIRTUAL_FIXED_GET_INTERFACE(33)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(34)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(34)
P11_VIRTUAL_FIXED_GET_INTERFACE(34)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(35)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(35)
P11_VIRTUAL_FIXED_GET_INTERFACE(35)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(36)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(36)
P11_VIRTUAL_FIXED_GET_INTERFACE(36)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(37)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(37)
P11_VIRTUAL_FIXED_GET_INTERFACE(37)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(38)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(38)
P11_VIRTUAL_FIXED_GET_INTERFACE(38)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(39)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(39)
P11_VIRTUAL_FIXED_GET_INTERFACE(39)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(40)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(40)
P11_VIRTUAL_FIXED_GET_INTERFACE(40)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(41)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(41)
P11_VIRTUAL_FIXED_GET_INTERFACE(41)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(42)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(42)
P11_VIRTUAL_FIXED_GET_INTERFACE(42)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(43)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(43)
P11_VIRTUAL_FIXED_GET_INTERFACE(43)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(44)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(44)
P11_VIRTUAL_FIXED_GET_INTERFACE(44)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(45)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(45)
P11_VIRTUAL_FIXED_GET_INTERFACE(45)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(46)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(46)
P11_VIRTUAL_FIXED_GET_INTERFACE(46)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(47)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(47)
P11_VIRTUAL_FIXED_GET_INTERFACE(47)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(48)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(48)
P11_VIRTUAL_FIXED_GET_INTERFACE(48)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(49)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(49)
P11_VIRTUAL_FIXED_GET_INTERFACE(49)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(50)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(50)
P11_VIRTUAL_FIXED_GET_INTERFACE(50)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(51)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(51)
P11_VIRTUAL_FIXED_GET_INTERFACE(51)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(52)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(52)
P11_VIRTUAL_FIXED_GET_INTERFACE(52)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(53)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(53)
P11_VIRTUAL_FIXED_GET_INTERFACE(53)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(54)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(54)
P11_VIRTUAL_FIXED_GET_INTERFACE(54)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(55)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(55)
P11_VIRTUAL_FIXED_GET_INTERFACE(55)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(56)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(56)
P11_VIRTUAL_FIXED_GET_INTERFACE(56)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(57)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(57)
P11_VIRTUAL_FIXED_GET_INTERFACE(57)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(58)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(58)
P11_VIRTUAL_FIXED_GET_INTERFACE(58)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(59)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(59)
P11_VIRTUAL_FIXED_GET_INTERFACE(59)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(60)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(60)
P11_VIRTUAL_FIXED_GET_INTERFACE(60)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(61)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(61)
P11_VIRTUAL_FIXED_GET_INTERFACE(61)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(62)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(62)
P11_VIRTUAL_FIXED_GET_INTERFACE(62)
P11_VIRTUAL_FIXED_GET_FUNCTION_LIST(63)
P11_VIRTUAL_FIXED_GET_INTERFACE_LIST(63)
P11_VIRTUAL_FIXED_GET_INTERFACE(63)
